# Pytorch with Distributed Deep Learning Tutorial

## Overview

This tutorial explains the necessary steps for enabling Distributed Deep Learning (DDL) from within the pytorch script examples provided in the PowerAI distribution.
DDL is directly integrated into Pytorch's distributed communication package, `torch.distributed`, as the backend `ddl`.

Additional information regarding `torch.distributed` :
[https://pytorch.org/docs/stable/distributed.html](https://pytorch.org/docs/stable/distributed.html)

## Enabling DDL in a Pytorch program

The DDL Pytorch integration makes it simple to run a Pytorch program on a cluster. To enable DDL, a user simply needs to initialize the Pytorch package `torch.distributed` with backend `ddl` before any other method in the program. The `init_method` needs to be set to `env://`.

```python
torch.distributed.init_process_group('ddl', init_method='env://')
```

For the distributed program to work most efficiently, the data needs to be split among each DDL instance. The following two methods are approaches in which the data can be split.

* Splitting Training Data Manually
* Splitting Training Data through Pytorch Modules `DistributedDataParallel` & `DistributedSampler`

## Splitting Training Data Manually

In this approach the data is split by the user, and the `dist.all_reduce` call is made manually.

The mnist.py script was modified in this manner to split the training data, as described above, to enable DDL. The modified script can be found in `$CONDA_PREFIX/lib/python$PY_VER/site-packages/torch/examples/ddl_examples/mnist/mnist.py`.

For more detailed information on how the data is being split, see Pytorch's tutorial on [Writing Distributed Applications](https://pytorch.org/tutorials/intermediate/dist_tuto.html)

## Splitting Training Data through Pytorch Module `DistributedDataParallel` & `DistributedSampler`

Pytorch offers a `DistributedSampler` module that performs the Training Data split amongst the DDL instances and `DistributedDataParallel` that does the averaging of the gradients on the backward pass.
DDL does not support the `num_workers` argument passed to `torch.utils.data.DataLoader`, unexpected behaviour may occur.

```python
kwargs = {'pin_memory': True}
train_sampler = torch.utils.data.distributed.DistributedSampler(dataset, world_size, rank)
train_data = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=False,
                                         sampler=train_sampler, **kwargs)
```

Then user simply needs to wrap their model with the `DistributedDataParallel` module.

```python
model = torch.nn.parallel.DistributedDataParallel(model)
```

The mnist_ddp.py script was modified in this manner to split the training data, as described above, to enable DDL. The modified script can be found in `$CONDA_PREFIX/lib/python$PY_VER/site-packages/torch/examples/ddl_examples/mnist/mnist_ddp.py`.

### Run Script Using ddlrun

DDL enabled programs should be launched using the `ddlrun` tool. See `$CONDA_PREFIX/doc/ddl/README.md` and `ddlrun -h` for more information on `ddlrun`.

Example of how to call `ddlrun`:
`ddlrun -H host1,host2 --accelerators 4 python $CONDA_PREFIX/lib/python$PY_VER/site-packages/torch/examples/ddl_examples/mnist/mnist.py --dist-backend ddl`

For additional information on how to use `ddlrun` see `$CONDA_PREFIX/doc/ddl/README.md`.

### Run DDL Enabled Imagenet script

Pytorch's imagenet script has been modified to take advtange of DDL.

Example of how to call DDL Enabled Imagenet:
`ddlrun --accelerators 4 python -u $CONDA_PREFIX/lib/python$PY_VER/site-packages/torch/examples/ddl_examples/imagenet/imagenet_ddp.py --dist-backend ddl --seed 1234 --dist-url env:// /dir/to/training_data`

For additional information regarding DDL Enabled Imagenet see:
`$CONDA_PREFIX/lib/python$PY_VER/site-packages/torch/examples/ddl_examples/imagenet/README.md`


&copy; Copyright IBM Corporation 2018
