"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""

import os

import ddlrun.utilities as utilities
import ddlrun.error as error


def mpirun_test():
    """Check to see if mpirun is on the path and is executable.

    Args:
        args: Parsed program arguments.

    Returns:
        None

    Raises:
        GenericDDLError
    """
    ret_code, mpipath, _ = utilities.run_command_capture("which mpirun")
    mpipath = mpipath.strip()
    if not ret_code or not (os.path.isfile(mpipath) and os.access(mpipath, os.X_OK)):
        raise error.GenericDDLError(error.ErrorMessage.NO_MPIRUN)
    return mpipath


def check_spectrum():
    """Determines if spectrum is the version of mpi installed.

    Returns:
        True if spectrum is installed, False otherwise.
    """
    _, mpi_version, _ = utilities.run_command_capture("mpirun -V")
    return "spectrum" in mpi_version


def mpi_test(args):
    """Checks to see if the executable 'mpitool' exists.

    -If it does, it is used to test the mpi arguments with the 'hello world' option.
    -Otherwise a warning is displayed, but execution continues.

    Args:
        args: Parsed program arguments.

    Returns:
        None

    Raises:
        GenericDDLError
    """
    # Only spectrum has mpitool, so skip this check if not using spectrum.
    if not args.use_spectrum:
        return

    # If mpitool exists and is executable, use that to test the specified MPI
    # options by running the mpitool "hello world" test.
    ret_code, stdout, stderr = utilities.mpirun_command(
        "which mpitool", conda_env=args.conda_env, args=args)
    if not ret_code:
        error.log_info(stdout)
        error.log_info(stderr)
        error.log_warning(error.ErrorMessage.NO_MPITOOL)
    else:
        ret_code, result, stderr = utilities.mpirun_command(
            "mpitool -hw", conda_env=args.conda_env, args=args)
        error.log_info(result)
        error.log_info(stderr)
        if not ret_code:
            raise error.GenericDDLError(error.ErrorMessage.MPI_OPS)


def verify_host_connectivity(args):
    """Iterates through the list of hosts, first checking that each host is
        reachable via mpi. If it is not, attempts to ssh to that host to
        get a better error message. Every host will be checked, even if an
        error is encountered.

    Args:
        args: Parsed program arguments.

    Returns:
        None

    Raises:
        DDLError
    """
    encountered_error = False
    for host in args.hosts:
        ret_code, result, stderr = utilities.mpirun_command(
            "true", host, args=args)
        error.log_info(result)
        error.log_info(stderr)
        if not ret_code:
            encountered_error = True
            error.log_error(error.ErrorMessage.INVALID_MPI_OPS, host)
            error.log_info(result)
            error.log_info(stderr)

            ret_code, result, stderr = utilities.run_command_capture(
                "ssh %s %s exit" % (utilities.make_arg_list(utilities.SSH_ARGS), host))
            error.log_info(result)
            error.log_info(stderr)
            if not ret_code:
                encountered_error = True
                error.log_error(error.ErrorMessage.SSH_ERROR, host)
                continue

    if encountered_error:
        raise error.DDLError()
