"""
*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
"""

import os
import re

import ddlrun.error as error
import ddlrun.utilities as utilities


def rankfile_test(args):
    """Tests that the rankfile exists and is accessible.

    Args:
        args: Parsed program arguments.

    Returns:
        None

    Raises:
        GenericDDLError
    """
    if args.rankfile is not None and not os.path.isfile(args.rankfile):
        raise error.GenericDDLError(error.ErrorMessage.NO_RANKFILE,
                                    args.rankfile)


def generate_rank_file(args):
    """Uses the specified rank gen arguments to generate a rankfile in the specified
        location.

    Args:
        args: Parsed program arguments.

    Returns:
        None

    Raises:
        GenericDDLError
    """
    ret_code, result, stderr = utilities.mpirun_command(
        "rank_gen.py %s %s" % (utilities.make_arg_list(args.rank_args),
                               ",".join(args.hosts)),
        args.hosts[0], conda_env=args.conda_env, args=args)
    error.log_info(result)
    error.log_info(stderr)

    if not ret_code:
        raise error.GenericDDLError(error.ErrorMessage.RANK_FILE_ERROR)
    with open(args.rankfile, "w") as rank_file:
        rank_file.write(result)


def parse_topology_from_rankfile(args, rankfile=None):
    """Parses the rankfile to construct the topology string.

    Args:
        args: Parsed program arguments.

    Returns:
        Topology String

    Raises:
        GenericDDLError
    """
    if rankfile is None:
        rankfile = utilities.read_all_text(args.rankfile)
    if not ("#aisles" in rankfile and "#racks" in rankfile
            and "#accelerators" in rankfile and "#host" in rankfile):
        error.log_info(rankfile)
        raise error.GenericDDLError(
            error.ErrorMessage.NO_AUTO_RANKFILE, args.rankfile,
            "Could not find the topology tags. (#aisles, #racks, etc.)")

    try:
        aisle_count = re.search(r"#aisles\s*=\s*(\d+)\s*", rankfile).group(1)
        rack_count = re.search(r"#racks\s*=\s*(\d+)\s*", rankfile).group(1)
        accel_count = re.search(r"#accelerators\s*=\s*(\d+)\s*",
                                rankfile).group(1)
        host_count = len(
            re.search(r"#host\s*=\s*(.+)\s*", rankfile).group(1).split(","))

        host_count = host_count // int(rack_count) // int(aisle_count)
    except:
        error.log_info(rankfile)
        raise error.GenericDDLError(
            error.ErrorMessage.NO_AUTO_RANKFILE, args.rankfile,
            "Could not parse one of the topology tags. (#aisles, #racks, etc.)"
        )

    if host_count <= 0:
        error.log_info(rankfile)
        raise error.GenericDDLError(error.ErrorMessage.INVALID_TOPO,
                                    args.rankfile)

    return "%sx%dx%sx%s" % (accel_count, host_count, rack_count, aisle_count)


def parse_hosts_from_rankfile(args):
    """Parses the rankfile to construct list of hosts with ports

    Args:
        args: Parsed program arguments.

    Returns:
        Hosts with port list

    Raises:
        None
    """
    if args.rankfile is None:
        host_list = "localhost:0"
        for i in range(1, int(args.num_procs)):
            host_list += "," + str(i)
        return host_list

    rankfile_content = utilities.read_all_text(args.rankfile)
    #remove block comments
    rankfile_content = re.sub(re.compile(r"/\*.*\*/"), "", rankfile_content)
    #remove comments
    rankfile_content = re.sub(
        re.compile(r"(//|#).*", re.MULTILINE), "", rankfile_content)

    host_ranks = dict()
    rank_regex = re.compile(r"^\s*rank\s+([0-9]+)\s*=\s*(\S+)")
    for line in rankfile_content.split("\n"):
        result = rank_regex.match(line.strip())
        if result is not None:
            rank = result.group(1)
            host = result.group(2)
            if not host in host_ranks.keys():
                host_ranks[host] = rank + ","
            else:
                host_ranks[host] = host_ranks[host] + rank + ","
    host_list = ""
    for key in host_ranks:
        ranks = host_ranks[key]
        host_list = host_list + key + ":" + ranks[0:-1] + ";"

    return host_list[0:-1]


def parse_hostfile(args):
    """Parse the host file and handle errors.

    Args:
        args: Parsed program arguments.

    Returns:
        Comma-delimitted host list.

    Raises:
        GenericDDLError
    """
    if args.host_file is None or not os.path.isfile(args.host_file):
        raise error.GenericDDLError(error.ErrorMessage.NO_HOSTFILE,
                                    args.host_file)

    host_file_contents = utilities.read_all_text(args.host_file)

    # Ignore comments.
    host_file_contents = re.sub(r"#.*", "", host_file_contents)
    # Trim whitespace.
    host_file_contents = "\n".join(
        line.strip() for line in re.split("\n+", host_file_contents))

    #Check for lines with commas/spaces to give better error message.
    if " " in host_file_contents or "," in host_file_contents:
        raise error.GenericDDLError(error.ErrorMessage.HOSTFILE_COMMA_SPACE,
                                    args.host_file)

    # Delimit lines with commas, ignoring empty lines.
    host_file_contents = ",".join(re.split("\n+", host_file_contents))

    # Ignore empty lines at beginning and end.
    host_file_contents = re.sub("(^,+)|(,+$)", "", host_file_contents)

    return host_file_contents


def parse_lsf_hosts():
    """Parse the hosts from the LSF environment variable 'LSB_MCPU_HOSTS'

    Returns:
        Comma-delimitted host list.

    Raises:
        GenericDDLError
    """
    hosts = utilities.LSFUtils.get_compute_hosts()
    return ",".join(hosts)


def topology_to_num_procs(topology):
    """Figure out the numprocs value based on topology.

    Args:
        topology: DDL topology string

    Returns:
        Total number of processes for given topology.

    Raises:
        None
    """
    running_prod = 1
    for val in topology.split("x"):
        running_prod *= int(val)
    return str(running_prod)


def get_topology(args):
    """Get the topology from the rankfile or from local specs.

    Args:
        args: Parsed input arguments.

    Returns:
        The topology string.
    """
    if args.rankfile is not None and args.topology is None:
        return parse_topology_from_rankfile(args)
    if args.topology is None:
        ret_code, result, stderr = utilities.run_command_capture(
            "rank_gen.py %s localhost" % (utilities.make_arg_list(
                args.rank_args)))
        if not ret_code:
            error.log_warning(error.ErrorMessage.HW_SPEC_ERROR)
            error.log_info(stderr)
            return "1x1"
        return parse_topology_from_rankfile(args, rankfile=result)
    return args.topology
